# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.



import pytest
import copy
import nbformat

from nbdime.diff_format import op_patch, op_addrange, op_removerange, op_replace
from .utils import sources_to_notebook, outputs_to_notebook, have_git, strip_cell_ids, new_cell_wo_id, deterministic_cell_ids
from nbdime.nbmergeapp import _build_arg_parser
from nbdime import merge_notebooks, apply_decisions
from nbdime.diffing.notebooks import diff_notebooks, set_notebook_diff_targets
from nbdime.merging.notebooks import decide_merge_with_diff, Strategies

# FIXME: Extend tests to more merge situations!


# Setup default args for merge app
builder = _build_arg_parser()
args = builder.parse_args(["", "", ""])


def test_merge_matching_notebooks(matching_nb_triplets, reset_log):
    "Test merge on pairs of notebooks with the same basename in the test suite."
    base, local, remote = matching_nb_triplets
    merge_notebooks(base, local, remote)
    # We can't really automate a generic merge test,
    # at least passing through code here...


def test_autoresolve_notebook_ec():
    # We need a source here otherwise the cells are not aligned
    source = "def foo(x, y):\n    return x**y"

    base = nbformat.v4.new_notebook()
    base["cells"].append(nbformat.v4.new_code_cell())
    base["cells"][0]["source"] = source
    base["cells"][0]["execution_count"] = 1

    local = copy.deepcopy(base)
    remote = copy.deepcopy(base)
    expected = copy.deepcopy(base)
    local["cells"][0]["execution_count"] = 2
    remote["cells"][0]["execution_count"] = 3
    expected["cells"][0]["execution_count"] = None

    merged, decisions = merge_notebooks(base, local, remote, args)

    assert merged == expected
    assert not any(d.conflict for d in decisions)


def test_merge_cell_sources_neighbouring_inserts_mergetool():
    base = strip_cell_ids(sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ]))
    local = strip_cell_ids(sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "print(f(3))",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ]))
    remote = strip_cell_ids(sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "print(f(7))",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ]))
    expected_partial = base
    expected_conflicts = [{
        "common_path": ("cells",),
        "local_diff": [op_addrange(1, [local.cells[1]])],
        "remote_diff": [op_addrange(1, [remote.cells[1]])]
    }]

    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"

    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, ignore_cell_ids=True)


def test_merge_cell_sources_neighbouring_inserts_union():
    base = strip_cell_ids(sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ]))
    local = strip_cell_ids(sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "print(f(3))",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ]))
    remote = strip_cell_ids(sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "print(f(7))",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ]))
    expected_partial = sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "print(f(3))",
        ], [
        "print(f(7))",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ])
    expected_conflicts = []

    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "union"

    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, ignore_cell_ids=True)


def test_merge_cell_sources_separate_inserts():
    base = strip_cell_ids(sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ]))
    local = strip_cell_ids(sources_to_notebook([[
        "print(f(3))",
        ], [
        "def f(x):\n",
        "    return x**2",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ],
        ]))
    remote = strip_cell_ids(sources_to_notebook([[
        "def f(x):\n",
        "    return x**2",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ], [
        "print(f(7))",
        ],
        ]))
    expected = strip_cell_ids(sources_to_notebook([[
        "print(f(3))",
        ], [
        "def f(x):\n",
        "    return x**2",
        ], [
        "def g(y):\n",
        "    return y + 2",
        ], [
        "print(f(7))",
        ],
        ]))
    actual, decisions = merge_notebooks(base, local, remote, args)
    assert not any([d.conflict for d in decisions])
    assert actual == expected


def src2nb(src, strip_ids=False):
    """Convert source strings to a notebook.

    src is either a single multiline string to become one cell,
    or a list (cells) of lists (lines) of singleline strings.
    """
    if isinstance(src, str):
        src = [[src]]
    if isinstance(src, list):
        src = sources_to_notebook(src)
    assert isinstance(src, dict)
    assert "cells" in src
    if strip_ids:
        strip_cell_ids(src)
    else:
        deterministic_cell_ids(src)
    return src


def _check(partial, expected_partial, decisions, expected_conflicts):
    sources = [cell.pop("source") for cell in partial["cells"]]
    expected_sources = [cell.pop("source") for cell in expected_partial["cells"]]
    assert sources == expected_sources

    outputs = [cell.pop("outputs") for cell in partial["cells"]]
    expected_outputs = [cell.pop("outputs") for cell in expected_partial["cells"]]
    assert outputs == expected_outputs

    assert partial == expected_partial
    conflicts = [d for d in decisions if d.conflict]
    expected_conflicts = copy.copy(expected_conflicts)
    assert len(conflicts) == len(expected_conflicts)
    for e, d in zip(expected_conflicts, conflicts):
        # Only check keys specified in expectation value
        for k in sorted(e.keys()):
            assert d[k] == e[k]


def _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args=None, ignore_cell_ids=False):
    base = src2nb(base, strip_ids=ignore_cell_ids)
    local = src2nb(local, strip_ids=ignore_cell_ids)
    remote = src2nb(remote, strip_ids=ignore_cell_ids)
    expected_partial = src2nb(expected_partial, strip_ids=ignore_cell_ids)
    merge_args = merge_args or args

    partial, decisions = merge_notebooks(base, local, remote, merge_args)

    _check(partial, expected_partial, decisions, expected_conflicts)


def _check_outputs(base, local, remote, expected_partial, expected_conflicts, merge_args=None, ignore_cell_ids=False):
    base = outputs_to_notebook(base, strip_ids=ignore_cell_ids)
    local = outputs_to_notebook(local, strip_ids=ignore_cell_ids)
    remote = outputs_to_notebook(remote, strip_ids=ignore_cell_ids)
    expected_partial = outputs_to_notebook(expected_partial, strip_ids=ignore_cell_ids)
    merge_args = merge_args or args

    partial, decisions = merge_notebooks(base, local, remote, merge_args)

    _check(partial, expected_partial, decisions, expected_conflicts)


def test_merge_simple_cell_source_no_change():
    # A very basic test: Just checking changes to a single cell source,

    # No change
    local = [["same"]]
    base = [["same"]]
    remote = [["same"]]
    expected_partial = [["same"]]
    expected_conflicts = []
    _check_sources(base, local, remote, expected_partial, expected_conflicts, ignore_cell_ids=True)


def test_merge_simple_cell_source_remote_change():
    # One sided change
    local = [["same"]]
    base = [["same"]]
    remote = [["different"]]
    expected_partial = [["different"]]
    expected_conflicts = []
    _check_sources(base, local, remote, expected_partial, expected_conflicts, ignore_cell_ids=True)


def test_merge_simple_cell_source_local_change():
    # One sided change
    local = [["different"]]
    base = [["same"]]
    remote = [["same"]]
    expected_partial = [["different"]]
    expected_conflicts = []
    _check_sources(base, local, remote, expected_partial, expected_conflicts, ignore_cell_ids=True)


def test_merge_simple_cell_source_agreed_change():
    # Same change on both sides
    local = [["different"]]
    base = [["same"]]
    remote = [["different"]]
    expected_partial = [["different"]]
    expected_conflicts = []
    _check_sources(base, local, remote, expected_partial, expected_conflicts, ignore_cell_ids=True)


def test_merge_simple_cell_source_conflicting_edit_aligned():
    # Conflicting edit in first line of single cell
    local = [["local\n", "some other\n", "lines\n", "to align\n"]]
    base = [["base\n", "some other\n", "lines\n", "to align\n"]]
    remote = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial = [["base\n", "some other\n", "lines\n", "to align\n"]]
    expected_conflicts = [{
        "common_path": ("cells", 0, "source"),
        "local_diff": [
            op_addrange(
                0, local[0][0:1]),
            op_removerange(0, 1)],
        "remote_diff": [
            op_addrange(
                0, remote[0][0:1]),
            op_removerange(0, 1)]
        }]
    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"

    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, ignore_cell_ids=True)


def test_merge_simple_cell_source_conflicting_insert():
    # Cell inserts at same location but no other modifications:
    # should this be accepted?
    local = [["base"], ["local"]]
    base = [["base"]]
    remote = [["base"], ["remote"]]
    if 1:  # Treat as conflict
        expected_partial = [["base"]]
        expected_conflicts = [{
            "common_path": ("cells",),
            "local_diff": [op_addrange(
                1, [new_cell_wo_id(local[1][0])]),
            ],
            "remote_diff": [op_addrange(
                1, [new_cell_wo_id(remote[1][0])]),
            ]
        }]
    else:  # Treat as non-conflict (insert both)
        expected_partial = [["base"], ["local"], ["remote"]]
        expected_conflicts = []

    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"

    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, True)


def test_merge_multiline_cell_source_conflict():
    # Modifying cell on both sides interpreted as editing the original cell
    # Note: This only works with cell ids hinting that the cells are the same
    source = [
        "def foo(x, y):\n",
        "    z = x * y\n",
        "    return z\n",
        ]
    local = [source + ["local\n"] + [""]]
    base = [source + [""]]
    remote = [source + ["remote\n"] + [""]]

    le = op_addrange(3, ["local\n"])
    re = op_addrange(3, ["remote\n"])

    expected_partial = base
    expected_conflicts = [{
        "common_path": ("cells", 0, "source"),
        "local_diff": [le],
        "remote_diff": [re]
    }]
    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"

    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args)


def test_merge_insert_cells_around_conflicting_cell():
    # Modifying an original cell and inserting a new cell on both sides
    source = [
        "def foo(x, y):\n",
        "    z = x * y\n",
        "    return z\n",
        ]
    local = [["new local cell\n"],
             source + ["local\n"]]
    base = [source]
    remote = [source + ["remote\n"],
              ["new remote cell\n"]]
    # Use mergetool strategy:
    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"
    if 0:
        # This is how it would look if neither source or cell inserts resulted
        # in conflicts:
        expected_partial = [local[0],
                            source + ["local\n", "remote\n"],
                            remote[1]]
        expected_conflicts = []
    elif 1:
        # This is how it would look if source inserts but not cell inserts
        # resulted in conflicts:
        expected_partial = [local[0], source, remote[1]]
        expected_conflicts = [{
            "common_path": ("cells", 0, "source"),
            "local_diff": [op_addrange(len(source), ["local\n"])],
            "remote_diff": [op_addrange(len(source), ["remote\n"])]
        }]
    else:
        # In current behaviour:
        # - base cell 0 is aligned correctly (this is the notebook diff heuristics)
        # - conflicting edits of base cell 0 detected
        # - local insert before cell 0 is treated as part of conflict on base cell 0
        # - remote insert after cell 0 is not treated as part of conflict
        # FIXME: This may not be the exact behaviour we want:
        # - For source, we might want both inserts to be part of the conflict.
        #   (if so, fix in generic merge and chunk collection)
        # - For cells, we might want both inserts to be ok, they are separate new cells after all. (use autoresolve for this?)
        # - Figure out the best behaviour and make it happen!
        expected_partial = [source, remote[1]]
        expected_conflicts = [{
            "common_path": ("cells",),
            "local_diff": [
                op_addrange(0, [new_cell_wo_id(
                    source=["new local cell"])]),
                op_patch(0, [op_patch("source", [
                    op_addrange(len("".join(source)), "local\n")])]),
            ],
            "remote_diff": [op_patch(0, [op_patch('source', [
                op_addrange(len("".join(source)), "remote\n")])])]
        }]
    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, True)


def test_merge_interleave_cell_add_remove():
    # Interleaving cell inserts and deletes
    # no modifications = avoids heuristics
    local = [["local 1\n"],
             ["base 1\n"],
             ["local 2\n"],
             ["base 2\n"],
             ["local 3\n"],
             ["base 3\n"],
             ["base 4\n"]]
    base = [["base 1\n"],
            ["base 2\n"],
            ["base 3\n"],
            ["base 4\n"]]
    remote = [["remote 1\n"],
              ["base 1\n"],
              ["remote 2\n"],
              ["remote 3\n"],
              ["base 3\n"]]
    # Note: in this case "remote 3" is before "local 3" because it's lumped
    # together with "remote 2" in the diff and the insert of
    # ["remote 2", "remote 3"]starts before the removal of "base 2"
    expected_partial = [["local 1\n"],
                        ["remote 1\n"],
                        ["base 1\n"],
                        ["local 2\n"],
                        ["remote 2\n"],
                        ["remote 3\n"],
                        ["local 3\n"],
                        ["base 3\n"]]
    expected_conflicts = []

    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "union"

    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, ignore_cell_ids=True)


def test_merge_conflicts_get_diff_indices_shifted():
    # Trying to induce conflicts with shifting of diff indices
    source = [
        "def foo(x, y):\n",
        "    z = x * y\n",
        "    return z\n",
        ]
    local = [["same"], source + ["local\n"], ["different"]]
    base = [["same"], source + ["base\n"], ["same"]]
    remote = [["different"], source + ["remote\n"], ["same"]]
    expected_partial = [["different"],
                        source + ["base\n"],
                        ["different"]]
    expected_conflicts = [{
        "common_path": ('cells', 1, 'source'),
        "local_diff": [
            op_addrange(3, ["local\n"]),
            op_removerange(3, 1),
        ],
        "remote_diff": [
            op_addrange(3, ["remote\n"]),
            op_removerange(3, 1),
        ]
    }]

    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"

    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args)

    # Trying to induce conflicts with shifting of diff indices
    local = [["same"],
             source + ["long line with minor change left"],
             ["different"]]
    base = [["same"],
            source + ["long line with minor change"],
            ["same"]]
    remote = [["different"],
              source + ["long line with minor change right"],
              ["same"]]
    expected_partial = [["different"],
                        source + ["long line with minor change"],
                        ["different"]]
    expected_conflicts = [{
        "common_path": ('cells', 1, 'source', 3),
        "local_diff": [
            op_addrange(27, " left"),
        ],
        "remote_diff": [
            op_addrange(27, " right"),
        ]
    }]
    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args)


@pytest.mark.xfail
def test_merge_inserts_within_deleted_range():
    # Multiple inserts within a deleted range
    base = """
def f(x):
    return x**2

def g(x):
    return x + 2
"""

    # Insert foo and bar
    local = """
def foo(y):
    return y / 3

def f(x):
    return x**2

def bar(y):
    return y - 3

def g(x):
    return x + 2
"""

    remote = ""  # Delete all

    if 0:
        # This is quite optimistic and would require employing aggressive
        # attempts at automatic resolution beyond what git and meld do:
        expected_partial = """
def foo(y):
    return y / 3

def bar(y):
    return y - 3
"""
    else:
        expected_partial = base
        expected_conflicts = [{
            "common_path": ("cells",),
            "local_diff": [op_patch("cells", [op_addrange(0, [
                nbformat.v4.new_code_cell(local)]), op_removerange(0, 1)])],
            "remote_diff": [op_patch("cells", [op_addrange(0, [
                nbformat.v4.new_code_cell(remote)]), op_removerange(0, 1)])]
        }]
    _check_sources(base, local, remote, expected_partial, expected_conflicts)

    # Keep it failing
    assert False


def test_merge_input_strategy_local_source_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\n", "some other\n", "lines\n", "to align\n"]]
    base = [["base\n", "some other\n", "lines\n", "to align\n"]]
    remote = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial = [["local\n", "some other\n", "lines\n", "to align\n"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.input_strategy = "use-local"
    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, ignore_cell_ids=True)


def test_merge_input_strategy_remote_source_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\n", "some other\n", "lines\n", "to align\n"]]
    base = [["base\n", "some other\n", "lines\n", "to align\n"]]
    remote = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.input_strategy = "use-remote"
    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, ignore_cell_ids=True)


def test_merge_input_strategy_base_source_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\n", "some other\n", "lines\n", "to align\n"]]
    base = [["base\n", "some other\n", "lines\n", "to align\n"]]
    remote = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial = [["base\n", "some other\n", "lines\n", "to align\n"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.input_strategy = "use-base"
    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, ignore_cell_ids=True)


def test_merge_input_strategy_union_source_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\n", "some other\n", "lines\n", "to align\n"]]
    base = [["base\n", "some other\n", "lines\n", "to align\n"]]
    remote = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial = [["local\n", "remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.input_strategy = "union"
    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args)


def test_merge_input_strategy_inline_source_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\n", "some other\n", "lines\n", "to align\n"]]
    base = [["base\n", "some other\n", "lines\n", "to align\n"]]
    remote = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    # Ideal case:
    if have_git:
        expected_partial = [[
            "<<<<<<< local\n",
            "local\n",
            "=======\n",
            "remote\n",
            ">>>>>>> remote\n",
            "some other\n",
            "lines\n",
            "to align\n",
            ]]
    else:
        # Fallback is not very smart yet:
        expected_partial = [[
            "<<<<<<< local\n",
            "local\n",
            "some other\n",
            "lines\n",
            "to align\n",
            #'||||||| base\n',
            #'base\n',
            #'some other\n',
            #'lines\n',
            #'to align\n',
            "=======\n",
            "remote\n",
            "some other\n",
            "lines\n",
            "to align\n",
            ">>>>>>> remote",
            ]]
    expected_conflicts = [{
        "common_path": ("cells", 0, "source"),
        "local_diff": [
            op_addrange(0, local[0][0:1]),
            op_removerange(0, 1)
            ],
        "remote_diff": [
            op_addrange(0, remote[0][0:1]),
            op_removerange(0, 1)
            ],
        "custom_diff": [
            op_addrange(0, expected_partial[0]),
            op_removerange(0, len(base[0]))
            ],
        }]
    expected_conflicts = [{
        "common_path": ("cells", 0),
        "conflict": True,
        "action": "custom",
        "local_diff": [op_patch("source", [
            op_addrange(0, local[0][0:1]),
            op_removerange(0, 1)
            ])],
        "remote_diff": [op_patch("source", [
            op_addrange(0, remote[0][0:1]),
            op_removerange(0, 1)
            ])],
        "custom_diff": [op_replace("source", "".join(expected_partial[0]))],
        }]
    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "use-base"
    merge_args.input_strategy = "inline"
    _check_sources(base, local, remote, expected_partial, expected_conflicts, merge_args, ignore_cell_ids=True)


def test_merge_output_strategy_local_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    remote = [["remote\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_partial = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.output_strategy = "use-local"
    _check_outputs(base, local, remote, expected_partial, expected_conflicts, merge_args, True)


def test_merge_output_strategy_remote_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    remote = [["remote\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_partial = [["remote\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.output_strategy = "use-remote"
    _check_outputs(base, local, remote, expected_partial, expected_conflicts, merge_args, True)


def test_merge_output_strategy_base_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    remote = [["remote\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_partial = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.output_strategy = "use-base"
    _check_outputs(base, local, remote, expected_partial, expected_conflicts, merge_args, True)


def test_merge_output_strategy_union_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    remote = [["remote\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_partial = [["local\nremote\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.output_strategy = "union"
    _check_outputs(base, local, remote, expected_partial, expected_conflicts, merge_args, True)


def test_merge_output_strategy_clear_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    remote = [["remote\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_partial = [["output2", "output3"]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.output_strategy = "remove"
    _check_outputs(base, local, remote, expected_partial, expected_conflicts, merge_args, True)


def test_merge_output_strategy_clear_all_conflict():
    # Conflicting cell inserts at same location as removing old cell
    local = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    remote = [["remote\nsome other\nlines\nto align\n", "output2", "output3"]]
    expected_partial = [[]]
    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.output_strategy = "clear-all"
    _check_outputs(base, local, remote, expected_partial, expected_conflicts, merge_args, True)


# TODO: Make test for output_strategy == 'inline'


def _make_notebook_with_multi_conflicts(
        expected_partial_source,
        expected_partial_metadata,
        expected_partial_outputs,
        ):
    local_source = [["local\n", "some other\n", "lines\n", "to align\n"]]
    base_source = [["base\n", "some other\n", "lines\n", "to align\n"]]
    remote_source = [["remote\n", "some other\n", "lines\n", "to align\n"]]

    local_metadata = [{'myval': 'local'}]
    base_metadata = [{'myval': 'base'}]
    remote_metadata = [{'myval': 'remote'}]

    local_outputs = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base_outputs = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    remote_outputs = [["remote\nsome other\nlines\nto align\n", "output2", "output3"]]

    base = outputs_to_notebook(base_outputs)
    local = outputs_to_notebook(local_outputs)
    remote = outputs_to_notebook(remote_outputs)
    expected_partial = outputs_to_notebook(expected_partial_outputs)

    for i in range(len(base.cells)):
        base.cells[i].source = "".join(base_source[i])
        local.cells[i].source = "".join(local_source[i])
        remote.cells[i].source = "".join(remote_source[i])
        expected_partial.cells[i].source = "".join(expected_partial_source[i])

        base.cells[i].metadata = base_metadata[i]
        local.cells[i].metadata = local_metadata[i]
        remote.cells[i].metadata = remote_metadata[i]
        expected_partial.cells[i].metadata = expected_partial_metadata[i]

    return base, local, remote, expected_partial


def test_metadata_union_strategy_metadata():
    # Conflicting cell inserts at same location as removing old cell
    expected_partial_source = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial_metadata = [{'myval': 'localremote'}]
    expected_partial_outputs = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base, local, remote, expected_partial = _make_notebook_with_multi_conflicts(
        expected_partial_source, expected_partial_metadata, expected_partial_outputs
    )

    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "union"
    merge_args.input_strategy = "use-remote"
    merge_args.output_strategy = "use-local"

    partial, decisions = merge_notebooks(base, local, remote, merge_args)

    _check(partial, expected_partial, decisions, expected_conflicts)


def test_metadata_union_strategy_not_applied_immutable_on_dict():
    # Conflicting cell inserts at same location as removing old cell
    expected_partial_source = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial_metadata = [{'myval': 5}]
    expected_partial_outputs = [["local\nsome other\nlines\nto align\n", "output2", "output3"]]
    base, local, remote, expected_partial = _make_notebook_with_multi_conflicts(
        expected_partial_source, expected_partial_metadata, expected_partial_outputs
    )
    base.cells[0].metadata['myval'] = 5
    local.cells[0].metadata['myval'] = 22
    remote.cells[0].metadata['myval'] = 13

    expected_conflicts = [{
        'action': 'base',
        'common_path': ('cells', 0, 'metadata'),
        'conflict': True,
        'local_diff': [{'key': 'myval', 'op': 'replace', 'value': 22}],
        'remote_diff': [{'key': 'myval', 'op': 'replace', 'value': 13}]
    }]
    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "union"
    merge_args.input_strategy = "use-remote"
    merge_args.output_strategy = "use-local"

    partial, decisions = merge_notebooks(base, local, remote, merge_args)

    _check(partial, expected_partial, decisions, expected_conflicts)


def test_merge_mix_strategies():
    # Conflicting cell inserts at same location as removing old cell
    expected_partial_source = [["remote\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial_metadata = [{'myval': 'local'}]
    expected_partial_outputs = [["local\nremote\nsome other\nlines\nto align\n", "output2", "output3"]]
    base, local, remote, expected_partial = _make_notebook_with_multi_conflicts(
        expected_partial_source, expected_partial_metadata, expected_partial_outputs
    )

    expected_conflicts = []
    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "use-local"
    merge_args.input_strategy = "use-remote"
    merge_args.output_strategy = "union"

    partial, decisions = merge_notebooks(base, local, remote, merge_args)

    _check(partial, expected_partial, decisions, expected_conflicts)


def test_autoresolve_empty_strategies():
    """Check that a autoresolve works with empty strategies"""
    expected_partial_source = [["base\n", "some other\n", "lines\n", "to align\n"]]
    expected_partial_metadata = [{'myval': 'base'}]
    expected_partial_outputs = [["base\nsome other\nlines\nto align\n", "output2", "output3"]]
    base, local, remote, expected_partial = _make_notebook_with_multi_conflicts(
        expected_partial_source, expected_partial_metadata, expected_partial_outputs
    )
    strip_cell_ids(base)
    strip_cell_ids(local)
    strip_cell_ids(remote)
    strip_cell_ids(expected_partial)

    expected_conflicts = [
        {
            'action': 'base',
            'common_path': ('cells', 0, 'source'),
            'conflict': True,
            'local_diff': [{'key': 0, 'op': 'addrange', 'valuelist': ['local\n']},
                           {'key': 0, 'length': 1, 'op': 'removerange'}],
            'remote_diff': [{'key': 0, 'op': 'addrange', 'valuelist': ['remote\n']},
                            {'key': 0, 'length': 1, 'op': 'removerange'}]
        },
        {
            'action': 'base',
            'common_path': ('cells', 0, 'outputs', 0, 'data', 'text/plain'),
            'conflict': True,
            'local_diff': [{'key': 0, 'op': 'addrange', 'valuelist': ['local\n']},
                           {'key': 0, 'length': 1, 'op': 'removerange'}],
            'remote_diff': [{'key': 0, 'op': 'addrange', 'valuelist': ['remote\n']},
                            {'key': 0, 'length': 1, 'op': 'removerange'}]
        },
        {
            'action': 'base',
            'common_path': ('cells', 0, 'metadata', 'myval'),
            'conflict': True,
            'local_diff': [{'key': 0, 'op': 'addrange', 'valuelist': ['local']},
                           {'key': 0, 'length': 1, 'op': 'removerange'}],
            'remote_diff': [{'key': 0, 'op': 'addrange', 'valuelist': ['remote']},
                            {'key': 0, 'length': 1, 'op': 'removerange'}]
        }
    ]

    # Since we cannot pass directly a strategies object, include copy of relevant code:
    local_diffs = diff_notebooks(base, local)
    remote_diffs = diff_notebooks(base, remote)

    strategies = Strategies()
    decisions = decide_merge_with_diff(
        base, local, remote,
        local_diffs, remote_diffs,
        strategies)

    partial = apply_decisions(base, decisions)

    _check(partial, expected_partial, decisions, expected_conflicts)


def test_only_sources(db, reset_notebook_diff, reset_log):
    base = db["mixed-conflicts--1"]
    local = db["mixed-conflicts--2"]
    remote = db["mixed-conflicts--3"]
    set_notebook_diff_targets(True, False, False, False)

    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"

    partial, decisions = merge_notebooks(base, local, remote, merge_args)

    assert len(decisions) > 0
    for d in decisions:
        path = d['common_path']
        # Still have some decisions on cell root, so avoid with len == 2 check
        assert len(path) == 2 or path[2] == 'source'


def test_only_outputs(db, reset_notebook_diff):
    base = db["mixed-conflicts--1"]
    local = db["mixed-conflicts--2"]
    remote = db["mixed-conflicts--3"]
    set_notebook_diff_targets(False, True, False, False)

    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"

    partial, decisions = merge_notebooks(base, local, remote, merge_args)

    assert len(decisions) > 0
    for d in decisions:
        path = d['common_path']
        # Still have some decisions on cell root, so avoid with len == 2 check
        assert len(path) == 2 or path[2] == 'outputs'


def test_only_metadata(db, reset_notebook_diff):
    base = db["mixed-conflicts--1"]
    local = db["mixed-conflicts--2"]
    remote = db["mixed-conflicts--3"]
    set_notebook_diff_targets(False, False, False, True)

    merge_args = copy.deepcopy(args)
    merge_args.merge_strategy = "mergetool"

    partial, decisions = merge_notebooks(base, local, remote, merge_args)

    assert len(decisions) > 0
    for d in decisions:
        path = d['common_path']
        # Still have some decisions on cell root, so avoid with len == 2 check
        assert (
            len(path) == 2 or
            path[0] == 'metadata' or
            path[2] == 'metadata' or
            path[4] == 'metadata'
        )




#!/usr/bin/env python
"""A git difftool plugin for notebooks.

Uses nbdime to display diffs of notebooks instead of plain text diffs of JSON.
Other files are displayed via `git diff`.

Enable in your global git config with:

    git-nbdifftool config --enable [--global | --system]

Use with:

    git difftool [<commit> [<commit>]]
"""
import os
import sys
from subprocess import check_call, check_output, CalledProcessError

from nbdime.args import (
    add_generic_args, add_git_config_subcommand, add_filter_args, ConfigBackedParser,
    Path,
)
from nbdime.webapp import nbdifftool
from .filter_integration import apply_possible_filter


def enable(scope=None, set_default=False):
    """Enable nbdime git difftool"""
    cmd = ['git', 'config']
    if scope:
        assert scope in ('global', 'system'), 'invalid scope value: %r' % scope
        cmd.append('--' + scope)

    check_call(cmd + ['difftool.nbdime.cmd', 'git-nbdifftool diff "$LOCAL" "$REMOTE" "$BASE"'])
    if set_default:
        check_call(cmd + ['diff.guitool', 'nbdime'])

    # Common setting:
    check_call(cmd + ['difftool.prompt', 'false'])


def disable(scope=None, _=None):
    """Disable nbdime git difftool"""
    cmd = ['git', 'config']
    if scope:
        assert scope in ('global', 'system'), 'invalid scope value: %r' % scope
        cmd.append('--' + scope)
    try:
        tool = check_output(cmd + ['diff.guitool']).decode('utf8', 'replace').strip()
    except CalledProcessError:
        pass
    else:
        if tool == 'nbdime':
            try:
                check_call(cmd + ['--unset', 'diff.guitool'])
            except CalledProcessError:
                # already unset
                pass


def show_diff(before, after, opts):
    """Run the difftool

    If we are diffing a notebook, show the diff via nbdiff-web.
    Otherwise, call out to `git diff`.
    """
    if before.endswith('.ipynb') or after.endswith('ipynb'):
        return nbdifftool.main_parsed(opts)
    else:
        # Never returns
        os.execvp('git', ['git', 'diff', before, after])


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    import argparse
    parser = ConfigBackedParser('git-nbdifftool', description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    add_generic_args(parser)
    subparsers = parser.add_subparsers(dest='subcommand')

    diff_parser = subparsers.add_parser('diff',
        description="The actual entrypoint for the diff tool. Git will call this."
    )
    add_filter_args(diff_parser)
    nbdifftool.build_arg_parser(diff_parser)
    diff_parser.add_argument('path', type=Path)

    config = add_git_config_subcommand(subparsers,
        enable, disable,
        subparser_help="Configure git to use nbdime via `git difftool`",
        enable_help="enable nbdime difftool via git config",
        disable_help="disable nbdime difftool via git config")
    config.add_argument('--set-default', action='store_true', dest='set_default',
        help="set nbdime as default gui difftool"
    )

    opts = parser.parse_args(args)
    if opts.subcommand == 'diff':
        remote = opts.remote
        if opts.use_filter and remote:
            remote = apply_possible_filter(opts.path, remote)
        return show_diff(opts.local, remote, opts)
    elif opts.subcommand == 'config':
        opts.config_func(opts.scope, opts.set_default)
        return 0
    else:
        parser.print_help()
        return 1


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import json
import os
import sys

from .args import (
    add_generic_args, add_diff_args, process_diff_flags, resolve_diff_args,
    add_diff_cli_args, add_prettyprint_args, ConfigBackedParser,
    prettyprint_config_from_args,
    Path,
    )
from .diffing.notebooks import diff_notebooks
from .gitfiles import changed_notebooks, is_gitref
from .prettyprint import pretty_print_notebook_diff
from .utils import EXPLICIT_MISSING_FILE, read_notebook, setup_std_streams


_description = "Compute the difference between two Jupyter notebooks."


def main_diff(args):
    """Main handler of diff CLI"""
    output = getattr(args, 'out', None)
    process_diff_flags(args)
    base, remote, paths = resolve_diff_args(args)

    # Check if base/remote are gitrefs:
    if is_gitref(base) and is_gitref(remote):
        # We are asked to do a diff of git revisions:
        status = 0
        for fbase, fremote in changed_notebooks(base, remote, paths):
            status = _handle_diff(fbase, fremote, output, args)
            if status != 0:
                # Short-circuit on error in diff handling
                return status
        return status
    else:  # Not gitrefs:
        return _handle_diff(base, remote, output, args)


def _handle_diff(base, remote, output, args):
    """Handles diffs of files, either as filenames or file-like objects"""
    # Check that if args are filenames they either exist, or are
    # explicitly marked as missing (added/removed):
    for fn in (base, remote):
        if (isinstance(fn, str) and not os.path.exists(fn) and
                fn != EXPLICIT_MISSING_FILE):
            print("Missing file {}".format(fn))
            return 1
    # Both files cannot be missing
    assert not (base == EXPLICIT_MISSING_FILE and remote == EXPLICIT_MISSING_FILE), (
        'cannot diff %r against %r' % (base, remote))

    # Perform actual work:
    a = read_notebook(base, on_null='empty')
    b = read_notebook(remote, on_null='empty')

    d = diff_notebooks(a, b)

    # Output as JSON to file, or print to stdout:
    if output:
        with open(output, "w") as df:
            # Compact version:
            #json.dump(d, df)
            # Verbose version:
            json.dump(d, df, indent=2, separators=(",", ": "))
    else:
        # This printer is to keep the unit tests passing,
        # some tests capture output with capsys which doesn't
        # pick up on sys.stdout.write()
        class Printer:
            def write(self, text):
                print(text, end="")
        # This sets up what to ignore:
        config = prettyprint_config_from_args(args, out=Printer())
        # Separate out filenames:
        base_name = base if isinstance(base, str) else base.name
        remote_name = remote if isinstance(remote, str) else remote.name
        pretty_print_notebook_diff(base_name, remote_name, a, d, config)

    return 0


def _build_arg_parser(prog=None):
    """Creates an argument parser for the nbdiff command."""
    parser = ConfigBackedParser(
        description=_description,
        prog=prog,
        )
    add_generic_args(parser)
    add_diff_args(parser)
    add_diff_cli_args(parser)
    add_prettyprint_args(parser)

    parser.add_argument(
        "base", help="the base notebook filename OR base git-revision.",
        type=Path,
        nargs='?', default='HEAD',
    )
    parser.add_argument(
        "remote", help="the remote modified notebook filename OR remote git-revision.",
        type=Path,
        nargs='?', default=None,
    )
    parser.add_argument(
        "paths", help="filter diffs for git-revisions based on path",
        type=Path,
        nargs='*', default=None,
    )

    parser.add_argument(
        '--out',
        type=Path,
        default=None,
        help="if supplied, the diff is written to this file. "
             "Otherwise it is printed to the terminal.")

    return parser


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    setup_std_streams()
    arguments = _build_arg_parser().parse_args(args)
    return main_diff(arguments)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from collections import namedtuple
import datetime
from difflib import unified_diff
import hashlib
import io
import os
import pprint
import re
import shutil
from subprocess import Popen, PIPE
import sys
import tempfile

from shutil import which

import colorama

from .diff_format import NBDiffFormatError, DiffOp, op_patch
from .ignorables import diff_ignorables
from .patching import patch
from .utils import star_path, split_path, join_path
from .utils import as_text, as_text_lines
from .log import warning


# Indentation offset in pretty-print
IND = "  "

# Max line width used some placed in pretty-print
# TODO: Use this for line wrapping some places?
MAXWIDTH = 78

git_diff_print_cmd = 'git diff --no-index --color-words before after'
diff_print_cmd = 'diff before after'
git_mergefile_print_cmd = 'git merge-file -p local base remote'
diff3_print_cmd = 'diff3 -m local base remote'


DIFF_ENTRY_END = '\n'

ColoredConstants = namedtuple('ColoredConstants', (
    'KEEP',
    'REMOVE',
    'ADD',
    'INFO',
    'RESET',
))


col_const = {
    True: ColoredConstants(
        KEEP   = '{color}   '.format(color=''),
        REMOVE = '{color}-  '.format(color=colorama.Fore.RED),
        ADD    = '{color}+  '.format(color=colorama.Fore.GREEN),
        INFO   = '{color}## '.format(color=colorama.Fore.BLUE + colorama.Style.BRIGHT),
        RESET  = colorama.Style.RESET_ALL,
    ),

    False: ColoredConstants(
        KEEP   = '   ',
        REMOVE = '-  ',
        ADD    = '+  ',
        INFO   = '## ',
        RESET  = '',
    )
}


class PrettyPrintConfig:
    def __init__(
            self,
            out=sys.stdout,
            include=None,
            color_words=False,
            use_git = True,
            use_diff = True,
            use_color = True,
            language = None
            ):
        self.out = out
        if include is None:
            for key in diff_ignorables:
                setattr(self, key, True)
        else:
            for key in diff_ignorables:
                setattr(self, key, getattr(include, key, True))

        self.color_words = color_words

        self.use_git = use_git
        self.use_diff = use_diff
        self.use_color = use_color
        self.language = language

    def should_ignore_path(self, path):
        starred = star_path(split_path(path))
        if starred.startswith('/cells/*/source'):
            return not self.sources
        if starred.startswith('/cells/*/attachments'):
            return not self.attachments
        if starred.startswith('/cells/*/metadata') or starred.startswith('/metadata'):
            return not self.metadata
        if starred.startswith('/cells/*/id'):
            return not self.id
        if starred.startswith('/cells/*/outputs'):
            return (
                not self.outputs or
                (starred == '/cells/*/outputs/*/execution_count' and
                 not self.details))
        # Can check against '/cells/*/' since we've processed all other
        # sub-keys that we know about above.
        if starred.startswith('/cells/*/'):
            return not self.details
        if starred.startswith('/nbformat'):
            return not self.details
        return False

    @property
    def KEEP(self):
        return col_const[self.use_color].KEEP

    @property
    def REMOVE(self):
        return col_const[self.use_color].REMOVE

    @property
    def ADD(self):
        return col_const[self.use_color].ADD

    @property
    def INFO(self):
        return col_const[self.use_color].INFO

    @property
    def RESET(self):
        return col_const[self.use_color].RESET

DefaultConfig = PrettyPrintConfig()


def external_merge_render(cmd, b, l, r):
    b = as_text(b)
    l = as_text(l)
    r = as_text(r)
    td = tempfile.mkdtemp()
    try:
        with io.open(os.path.join(td, 'local'), 'w', encoding="utf8") as f:
            f.write(l)
        with io.open(os.path.join(td, 'base'), 'w', encoding="utf8") as f:
            f.write(b)
        with io.open(os.path.join(td, 'remote'), 'w', encoding="utf8") as f:
            f.write(r)
        assert all(fn in cmd for fn in ['local', 'base', 'remote']), (
            'invalid cmd argument for external merge renderer')
        p = Popen(cmd, cwd=td, stdout=PIPE)
        output, errors = p.communicate()
        status = p.returncode
        output = output.decode('utf8')
        # normalize newlines
        output = output.replace('\r\n', '\n')
    finally:
        shutil.rmtree(td)
    return output, status


def external_diff_render(cmd, a, b):
    a = as_text(a)
    b = as_text(b)
    td = tempfile.mkdtemp()
    try:
        # TODO: Pass in language information so that an appropriate file
        # extension can be used. This should provide a hint to the differ.
        with io.open(os.path.join(td, 'before'), 'w', encoding="utf8") as f:
            f.write(a)
        with io.open(os.path.join(td, 'after'), 'w', encoding="utf8") as f:
            f.write(b)
        assert all(fn in cmd for fn in ['before', 'after']), (
            'invalid cmd argument for external diff renderer: %r' %
            cmd)
        p = Popen(cmd, cwd=td, stdout=PIPE)
        output, errors = p.communicate()
        status = p.returncode
        output = output.decode('utf8')
        r = re.compile(r"^\\ No newline at end of file\n?", flags=re.M)
        output, n = r.subn("", output)
        assert n <= 2, 'unexpected output from external diff renderer'
    finally:
        shutil.rmtree(td)
    return output, status


def format_merge_render_lines(
        base, local, remote,
        base_title, local_title, remote_title,
        marker_size, include_base):
    sep0 = "<"*marker_size
    sep1 = "|"*marker_size
    sep2 = "="*marker_size
    sep3 = ">"*marker_size

    if local and local[-1].endswith('\n'):
        local[-1] = local[-1] + '\n'
    if remote and remote[-1].endswith('\n'):
        remote[-1] = remote[-1] + '\n'

    # Extract equal lines at beginning
    prelines = []
    i = 0
    n = min(len(local), len(remote))
    while i < n and local[i] == remote[i]:
        prelines.append(local[i])
        i += 1
    local = local[i:]
    remote = remote[i:]

    # Extract equal lines at end
    postlines = []
    i = len(local) - 1
    j = len(remote) - 1
    while (i >= 0 and i < len(local) and
           j >= 0 and j < len(remote) and
           local[i] == remote[j]):
        postlines.append(local[i])
        i += 1
        j += 1
    postlines = reversed(postlines)
    local = local[:i+1]
    remote = remote[:j+1]

    lines = []
    lines.extend(prelines)

    sep0 = "%s %s\n" % (sep0, local_title)
    lines.append(sep0)
    lines.extend(local)

    # This doesn't take prelines and postlines into account
    # if include_base:
    #     sep1 = "%s %s\n" % (sep1, base_title)
    #     lines.append(sep1)
    #     lines.extend(base)

    sep2 = "%s\n" % (sep2,)
    lines.append(sep2)
    lines.extend(remote)

    sep3 = "%s %s\n" % (sep3, remote_title)
    lines.append(sep3)

    lines.extend(postlines)

    # Make sure all but the last line ends with newline
    for i in range(len(lines)):
        if not lines[i].endswith('\n'):
            lines[i] = lines[i] + '\n'
    if lines:
        lines[-1] = lines[-1].rstrip("\r\n")

    return lines


def builtin_merge_render(base, local, remote, strategy=None):
    if local == remote:
        return local, 0

    # In this extremely simplified merge rendering,
    # we currently define conflict as local != remote

    if strategy == "use-local":
        return local, 0
    elif strategy == "use-remote":
        return remote, 0
    elif strategy is not None:
        warning("Using builtin merge render but ignoring strategy %s", strategy)

    # Styling
    local_title = "local"
    base_title = "base"
    remote_title = "remote"
    marker_size = 7  # git uses 7 by default

    include_base = False  # TODO: Make option

    local = as_text_lines(local)
    base = as_text_lines(base)
    remote = as_text_lines(remote)

    lines = format_merge_render_lines(
        base, local, remote,
        base_title, local_title, remote_title,
        marker_size, include_base
        )

    merged = "".join(lines)
    return merged, 1


def builtin_diff_render(a, b, config):
    gen = unified_diff(
        a.splitlines(False),
        b.splitlines(False),
        lineterm='')
    uni = []
    for line in gen:
        if line.startswith('+'):
            uni.append("%s%s%s" % (config.ADD, line[1:], config.RESET))
        elif line.startswith('-'):
            uni.append("%s%s%s" % (config.REMOVE, line[1:], config.RESET))
        elif line.startswith(' '):
            uni.append("%s%s%s" % (config.KEEP, line[1:], config.RESET))
        elif line.startswith('@'):
            uni.append(line)
        else:
            # Don't think this will happen?
            uni.append("%s%s%s" % (config.KEEP, line[1:], config.RESET))
    return '\n'.join(uni)


def diff_render_with_git(a, b, config):
    cmd = git_diff_print_cmd
    if not config.use_color:
        cmd = cmd.replace(" --color-words", "")
    elif not config.color_words:
        # Will do nothing if use_color is not True:
        cmd = cmd.replace("--color-words", "--color")
    diff, status = external_diff_render(cmd.split(), a, b)
    return "".join(diff.splitlines(True)[4:])


def diff_render_with_diff(a, b):
    cmd = diff_print_cmd
    diff, status = external_diff_render(cmd.split(), a, b)
    return diff


def diff_render_with_difflib(a, b, config):
    diff = builtin_diff_render(a, b, config)
    return "".join(diff.splitlines(True)[2:])


def diff_render(a, b, config=DefaultConfig):
    if config.use_git and which('git'):
        return diff_render_with_git(a, b, config)
    elif config.use_diff and which('diff'):
        return diff_render_with_diff(a, b)
    else:
        return diff_render_with_difflib(a, b, config)


def merge_render_with_git(b, l, r, strategy=None):
    # Note: git merge-file also takes argument -L to change label if needed
    cmd = git_mergefile_print_cmd
    if strategy == "use-local":
        cmd += " --ours"
    elif strategy == "use-remote":
        cmd += " --theirs"
    elif strategy == "union":
        cmd += " --union"
    elif strategy is not None:
        warning("Using git merge-file but ignoring strategy %s", strategy)
    merged, status = external_merge_render(cmd.split(), b, l, r)

    # Remove trailing newline if ">>>>>>> remote" is the last line
    lines = merged.splitlines(True)
    if "\n" in lines[-1] and (">"*7) in lines[-1]:
        merged = merged.rstrip()
    return merged, status


def merge_render_with_diff3(b, l, r, strategy=None):
    # Note: diff3 also takes argument -L to change label if needed
    cmd = diff3_print_cmd
    if strategy == "use-local":
        return l, 0
    elif strategy == "use-remote":
        return r, 0
    elif strategy is not None:
        warning("Using diff3 but ignoring strategy %s", strategy)
    merged, status = external_merge_render(cmd.split(), b, l, r)
    return merged, status


def merge_render(b, l, r, strategy=None, config=DefaultConfig):
    if strategy == "use-base":
        return b, 0
    if config.use_git and which('git'):
        return merge_render_with_git(b, l, r, strategy)
    elif config.use_diff and which('diff3'):
        return merge_render_with_diff3(b, l, r, strategy)
    else:
        return builtin_merge_render(b, l, r, strategy)


def file_timestamp(filename):
    "Return modification time for filename as a string."
    if os.path.exists(filename):
        t = os.path.getmtime(filename)
        dt = datetime.datetime.fromtimestamp(t)
        return dt.isoformat(str(" "))
    else:
        return "(no timestamp)"


def hash_string(s):
    return hashlib.md5(s.encode("utf8")).hexdigest()

_base64 = re.compile(
    r'^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$',
    re.MULTILINE | re.UNICODE)

def _trim_base64(s):
    """Trim and hash base64 strings"""
    if len(s) > 64 and _base64.match(s.replace('\n', '')):
        h = hash_string(s)
        s = '%s...<snip base64, md5=%s...>' % (s[:8], h[:16])
    return s


def format_value(v):
    "Format simple value for printing. Snips base64 strings and uses pprint for the rest."
    if not isinstance(v, str):
        # Not a string, defer to pprint
        vstr = pprint.pformat(v)
    else:
        # Snip if base64 data
        vstr = _trim_base64(v)
    return vstr


def pretty_print_value(value, prefix="", config=DefaultConfig):
    """Print a possibly complex value with all lines prefixed.

    Calls out to generic formatters based on value
    type for dicts, lists, and multiline strings.
    Uses format_value for simple values.
    """
    if isinstance(value, dict):
        pretty_print_dict(value, (), prefix, config)
    elif isinstance(value, list) and value:
        pretty_print_list(value, prefix, config)
    else:
        pretty_print_multiline(format_value(value), prefix, config)


def pretty_print_value_at(value, path, prefix="", config=DefaultConfig):
    """Print a possibly complex value with all lines prefixed.

    Calls out to other specialized formatters based on path
    for cells, outputs, attachments, and more generic formatters
    based on type for dicts, lists, and multiline strings.
    Uses format_value for simple values.
    """
    # Format starred version of path
    if path is None:
        starred = None
    else:
        if path.startswith('/'):
            path_prefix, path_trail = ('', path)
        else:
            path_prefix, path_trail = path.split('/', 1)
        starred = star_path(split_path(path_trail))

    # Check if we can handle path with specific formatter
    if starred is not None:
        if starred == "/cells/*":
            pretty_print_cell(None, value, prefix, True, config)
        elif starred == "/cells":
            for cell in value:
                pretty_print_cell(None, cell, prefix, True, config)
        elif starred == "/cells/*/outputs/*":
            pretty_print_output(None, value, prefix, config)
        elif starred == "/cells/*/outputs":
            for output in value:
                pretty_print_output(None, output, prefix, config)
        elif starred == "/cells/*/attachments":
            pretty_print_attachments(value, prefix, config)
        else:
            starred = None

    if starred is None:
        pretty_print_value(value, prefix, config)


def pretty_print_key(k, prefix, config):
    config.out.write("%s%s:\n" % (prefix, k))


def pretty_print_key_value(k, v, prefix, config):
    config.out.write("%s%s: %s\n" % (prefix, k, v))


def pretty_print_diff_action(msg, path, config):
    config.out.write("%s%s %s:%s\n" % (config.INFO, msg, path, config.RESET))


def pretty_print_item(k, v, prefix="", config=DefaultConfig):
    if isinstance(v, dict):
        pretty_print_key(k, prefix, config)
        pretty_print_dict(v, (), prefix+IND, config)
    elif isinstance(v, list):
        pretty_print_key(k, prefix, config)
        pretty_print_list(v, prefix+IND, config)
    else:
        vstr = format_value(v)
        if "\n" in vstr:
            # Multiline strings
            pretty_print_key(k, prefix, config)
            for line in vstr.splitlines(False):
                config.out.write("%s%s\n" % (prefix+IND, line))
        else:
            # Singleline strings
            pretty_print_key_value(k, vstr, prefix, config)


def pretty_print_multiline(text, prefix="", config=DefaultConfig):
    assert isinstance(text, str), 'expected string argument'

    # Preprend prefix to lines, letting lines keep their own newlines
    lines = text.splitlines(True)
    for line in lines:
        config.out.write(prefix + line)

    # If the final line doesn't have a newline,
    # make sure we still start a new line
    if not text.endswith("\n"):
        config.out.write("\n")


def pretty_print_list(li, prefix="", config=DefaultConfig):
    listr = pprint.pformat(li)
    if len(listr) < MAXWIDTH - len(prefix) and "\\n" not in listr:
        config.out.write("%s%s\n" % (prefix, listr))
    else:
        for k, v in enumerate(li):
            pretty_print_item("item[%d]" % k, v, prefix, config)


def pretty_print_dict(d, exclude_keys=(), prefix="", config=DefaultConfig):
    """Pretty-print a dict without wrapper keys

    Instead of {'key': 'value'}, do

        key: value
        key:
          long
          value

    """
    for k in sorted(set(d) - set(exclude_keys)):
        v = d[k]
        pretty_print_item(k, v, prefix, config)


def pretty_print_metadata(md, known_keys, prefix="", config=DefaultConfig):
    md1 = {}
    md2 = {}
    for k in md:
        if k in known_keys:
            md1[k] = md[k]
        else:
            md2[k] = md[k]
    if md1:
        pretty_print_key("metadata (known keys)", prefix, config)
        pretty_print_dict(md1, (), prefix+IND, config)
    if md2:
        pretty_print_key("metadata (unknown keys)", prefix, config)
        pretty_print_dict(md2, (), prefix+IND, config)


def pretty_print_output(i, output, prefix="", config=DefaultConfig):
    oprefix = prefix+IND
    numstr = "" if i is None else " %d" % i
    k = "output%s" % (numstr,)
    pretty_print_key(k, prefix, config)

    item_keys = ("output_type", "execution_count",
                 "name", "text", "data",
                 "ename", "evalue", "traceback")
    for k in item_keys:
        v = output.get(k)
        if v:
            pretty_print_item(k, v, oprefix, config)

    exclude_keys = {"output_type", "metadata", "traceback"} | set(item_keys)

    metadata = output.get("metadata")
    if metadata:
        known_output_metadata_keys = {"isolated"}
        pretty_print_metadata(metadata, known_output_metadata_keys, oprefix, config)

    pretty_print_dict(output, exclude_keys, oprefix, config)


def pretty_print_outputs(outputs, prefix="", config=DefaultConfig):
    pretty_print_key("outputs", prefix, config)
    for i, output in enumerate(outputs):
        pretty_print_output(i, output, prefix+IND, config)


def pretty_print_attachments(attachments, prefix="", config=DefaultConfig):
    pretty_print_key("attachments", prefix, config)
    for name in sorted(attachments):
        pretty_print_item(name, attachments[name], prefix+IND, config)

try:
    from pygments import highlight
    from pygments.formatters import Terminal256Formatter
    from pygments.lexers import find_lexer_class_by_name
    from pygments.util import ClassNotFound

    def colorize_source(source, lexer_name):
        try:
            lexer = find_lexer_class_by_name(lexer_name)()
        except ClassNotFound:
            return source
        formatter = Terminal256Formatter()
        return highlight(source, lexer, formatter)

except ImportError as e:
    def colorize_source(source, *args, **kwargs):
        return source


def pretty_print_source(source, prefix="", is_markdown=False, config=DefaultConfig):
    pretty_print_key("source", prefix, config)
    if not prefix.strip() and (is_markdown or config.language):
        source_highlighted = colorize_source(
            source,
            'markdown' if is_markdown else config.language
        )
    else:
        source_highlighted = source
    pretty_print_multiline(source_highlighted, prefix+IND, config)


def pretty_print_cell(i, cell, prefix="", force_header=False, config=DefaultConfig):
    key_prefix = prefix+IND

    def c():
        "Write cell header first time this is called."
        if not c.called:
            # Write cell type and optionally number:
            numstr = "" if i is None else " %d" % i
            k = "%s cell%s" % (cell.get("cell_type"), numstr)
            pretty_print_key(k, prefix, config)
            c.called = True
    c.called = False

    if force_header:
        c()

    id = cell.get("id")
    if id and config.details:
        # Write cell ID if there
        c()
        pretty_print_item("id", id, key_prefix, config)

    execution_count = cell.get("execution_count")
    if execution_count and config.details:
        # Write execution count if there (only source cells)
        c()
        pretty_print_item("execution_count", execution_count, key_prefix, config)

    metadata = cell.get("metadata")
    if metadata and config.metadata:
        # Write cell metadata
        c()
        known_cell_metadata_keys = {
            "collapsed", "autoscroll", "deletable", "format", "name", "tags",
        }
        pretty_print_metadata(
            cell.metadata,
            known_cell_metadata_keys,
            key_prefix,
            config)

    source = cell.get("source")
    if source and config.sources:
        is_markdown = cell.get('cell_type', None) == 'markdown'
        # Write source
        c()
        pretty_print_source(source, key_prefix, is_markdown=is_markdown, config=config)

    attachments = cell.get("attachments")
    if attachments and config.attachments:
        # Write attachment if there (only markdown and raw cells)
        c()
        pretty_print_attachments(attachments, key_prefix, config)

    outputs = cell.get("outputs")
    if outputs and config.outputs:
        # Write outputs if there (only source cells)
        c()
        pretty_print_outputs(outputs, key_prefix, config)

    exclude_keys = {
        'cell_type', 'source', 'execution_count', 'outputs', 'metadata',
        'id', 'attachment',
    }
    if (set(cell) - exclude_keys) and config.details:
        # present anything we haven't special-cased yet (future-proofing)
        c()
        pretty_print_dict(cell, exclude_keys, key_prefix, config)


def pretty_print_notebook(nb, config=DefaultConfig):
    """Pretty-print a notebook for debugging, skipping large details in metadata and output

    Parameters
    ----------

    nb: dict
        The notebook object
    config: PrettyPrintConfig
        A config object determining what is printed and where
    """
    prefix = ""

    if config.language is None:
        language_info = nb.metadata.get('language_info', {})
        config.language = language_info.get(
            'pygments_lexer',
            language_info.get('name', None)
        )

    if config.details:
        # Write notebook header
        v = "%d.%d" % (nb.nbformat, nb.nbformat_minor)
        pretty_print_key_value("notebook format", v, prefix, config)

    # Report unknown keys
    unknown_keys = set(nb.keys()) - {"nbformat", "nbformat_minor", "metadata", "cells"}
    if unknown_keys:
        pretty_print_key_value("unknown keys", repr(unknown_keys), prefix, config)

    if config.metadata:
        # Write notebook metadata
        known_metadata_keys = {"kernelspec", "language_info"}
        pretty_print_metadata(nb.metadata, known_metadata_keys, "", config)

    # Write notebook cells
    for i, cell in enumerate(nb.cells):
        pretty_print_cell(i, cell, prefix="", config=config)


def pretty_print_diff_entry(a, e, path, config=DefaultConfig):
    if config.should_ignore_path(path):
        return
    key = e.key
    nextpath = "/".join((path, str(key)))
    op = e.op

    # Recurse to handle patch ops
    if op == DiffOp.PATCH:
        # Useful for debugging:
        #if not (len(e.diff) == 1 and e.diff[0].op == DiffOp.PATCH):
        #    config.out.write("{}// patch -+{} //{}\n".format(INFO, nextpath, RESET))
        #else:
        #    config.out.write("{}// patch... -+{} //{}\n".format(INFO, nextpath, RESET))
        pretty_print_diff(a[key], e.diff, nextpath, config)
        return

    if op == DiffOp.ADDRANGE:
        pretty_print_diff_action("inserted before", nextpath, config)
        pretty_print_value_at(e.valuelist, path, config.ADD, config)

    elif op == DiffOp.REMOVERANGE:
        if e.length > 1:
            keyrange = "{}-{}".format(nextpath, key + e.length - 1)
        else:
            keyrange = nextpath
        pretty_print_diff_action("deleted", keyrange, config)
        pretty_print_value_at(a[key: key + e.length], path, config.REMOVE, config)

    elif op == DiffOp.REMOVE:
        if config.should_ignore_path(nextpath):
            return
        pretty_print_diff_action("deleted", nextpath, config)
        pretty_print_value_at(a[key], nextpath, config.REMOVE, config)

    elif op == DiffOp.ADD:
        if config.should_ignore_path(nextpath):
            return
        pretty_print_diff_action("added", nextpath, config)
        pretty_print_value_at(e.value, nextpath, config.ADD, config)

    elif op == DiffOp.REPLACE:
        if config.should_ignore_path(nextpath):
            return
        aval = a[key]
        bval = e.value
        if type(aval) is not type(bval):
            typechange = " (type changed from %s to %s)" % (
                aval.__class__.__name__, bval.__class__.__name__)
        else:
            typechange = ""
        pretty_print_diff_action("replaced" + typechange, nextpath, config)
        pretty_print_value_at(aval, nextpath, config.REMOVE, config)
        pretty_print_value_at(bval, nextpath, config.ADD, config)

    else:
        raise NBDiffFormatError("Unknown list diff op {}".format(op))

    config.out.write(DIFF_ENTRY_END + config.RESET)


def pretty_print_dict_diff(a, di, path, config=DefaultConfig):
    "Pretty-print a nbdime diff."
    for key, e in sorted([(e.key, e) for e in di], key=lambda x: x[0]):
        pretty_print_diff_entry(a, e, path, config)


def pretty_print_list_diff(a, di, path, config=DefaultConfig):
    "Pretty-print a nbdime diff."
    for e in di:
        pretty_print_diff_entry(a, e, path, config)


def pretty_print_string_diff(a, di, path, config=DefaultConfig):
    "Pretty-print a nbdime diff."
    pretty_print_diff_action("modified", path, config)

    b = patch(a, di)

    ta = _trim_base64(a)
    tb = _trim_base64(b)

    if ta != a or tb != b:
        if ta != a:
            config.out.write('%s%s\n' % (config.REMOVE, ta))
        else:
            pretty_print_value_at(a, path, config.REMOVE, config)
        if tb != b:
            config.out.write('%s%s\n' % (config.ADD, tb))
        else:
            pretty_print_value_at(b, path, config.ADD, config)
    elif "\n" in a or "\n" in b:
        # Delegate multiline diff formatting
        diff = diff_render(a, b, config)
        config.out.write(diff)
    else:
        # Just show simple -+ single line (usually metadata values etc)
        config.out.write("%s%s\n" % (config.REMOVE, a))
        config.out.write("%s%s\n" % (config.ADD, b))

    config.out.write(DIFF_ENTRY_END + config.RESET)


def pretty_print_diff(a, di, path, config=DefaultConfig):
    "Pretty-print a nbdime diff."
    if isinstance(a, dict):
        pretty_print_dict_diff(a, di, path, config)
    elif isinstance(a, list):
        pretty_print_list_diff(a, di, path, config)
    elif isinstance(a, str):
        pretty_print_string_diff(a, di, path, config)
    else:
        raise NBDiffFormatError(
            "Invalid type {} for diff presentation.".format(type(a))
        )


notebook_diff_header = """\
nbdiff {afn} {bfn}
--- {afn}{atime}
+++ {bfn}{btime}
"""

def pretty_print_notebook_diff(afn, bfn, a, di, config=DefaultConfig):
    """Pretty-print a notebook diff

    Parameters
    ----------

    afn: str
        Filename of a, the base notebook
    bfn: str
        Filename of b, the updated notebook
    a: dict
        The base notebook object
    di: diff
        The diff object describing the transformation from a to b
    config: PrettyPrintConfig
        Config object determining what gets printed and where
    """
    if di:
        path = ""
        atime = "  " + file_timestamp(afn)
        btime = "  " + file_timestamp(bfn)
        config.out.write(notebook_diff_header.format(
            afn=afn, bfn=bfn, atime=atime, btime=btime))
        pretty_print_diff(a, di, path, config)


def pretty_print_merge_decision(base, decision, config=DefaultConfig):
    prefix = IND

    path = join_path(decision.common_path)
    confnote = "conflicted " if decision.conflict else ""
    config.out.write("%s%sdecision at %s:%s\n" % (
        config.INFO.replace("##", "===="), confnote, path, config.RESET))

    diff_keys = ("diff", "local_diff", "remote_diff", "custom_diff", "similar_insert")
    exclude_keys = set(diff_keys) | {"common_path", "action", "conflict"}
    pretty_print_dict(decision, exclude_keys, prefix, config)

    for dkey in diff_keys:
        diff = decision.get(dkey)

        if (dkey == "remote_diff" and decision.action == "either" and
                diff == decision.get("local_diff")):
            # Skip remote diff
            continue
        elif (dkey == "local_diff" and decision.action == "either" and
                diff == decision.get("remote_diff")):
            note = " (same as remote_diff)"
        elif dkey.startswith(decision.action):
            note = " (selected)"
        else:
            note = ""

        if diff:
            config.out.write("%s%s%s:%s\n" % (
                config.INFO.replace("##", "---"), dkey, note, config.RESET))
            value = base
            for i, k in enumerate(decision.common_path):
                if isinstance(value, str):
                    # Example case:
                    #   common_path = /cells/0/source/3
                    #   value = nb.cells[0].source
                    #   k = line number 3
                    #   k is last item in common_path
                    assert i == len(decision.common_path) - 1, (
                        'invalid discision common path, tries to subindex string: %r' %
                        decision.common_path)

                    # Diffs on strings are usually line-based, _except_
                    # when common_path points to a line within a string.
                    # Wrap character based diff in a patch op with line
                    # number to normalize.
                    diff = [op_patch(k, diff)]
                    break
                else:
                    # Either a list or dict, get subvalue
                    value = value[k]
            pretty_print_diff(value, diff, path, config)


#def pretty_print_string_diff(string, lineno, diff, config):
#    line = string.splitlines(True)[lineno]
#    pretty_print_diff_entry(e)


def pretty_print_merge_decisions(base, decisions, config=DefaultConfig):
    """Pretty-print notebook merge decisions

    Parameters
    ----------

    base: dict
        The base notebook object
    decisions: list
        The list of merge decisions
    """
    conflicted = [d for d in decisions if d.conflict]
    config.out.write("%d conflicted decisions of %d total:\n"
              % (len(conflicted), len(decisions)))
    for d in decisions:
        pretty_print_merge_decision(base, d, config)


def pretty_print_notebook_merge(bfn, lfn, rfn, bnb, lnb, rnb, mnb, decisions, config=DefaultConfig):
    """Pretty-print a notebook merge

    Parameters
    ----------

    bfn: str
        Filename of the base notebook
    lfn: str
        Filename of the local notebook
    rfn: str
        Filename of the remote notebook
    bnb: dict
        The base notebook object
    lnb: dict
        The local notebook object
    rnb: dict
        The remote notebook object
    mnb: dict
        The partially merged notebook object
    decisions: list
        The list of merge decisions including conflicts
    """
    pretty_print_merge_decisions(bnb, decisions, config)



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""
Utilities for computing 'snakes', or contiguous sequences of equal elements of two sequences.
"""

from ..diff_format import SequenceDiffBuilder
from .seq_bruteforce import bruteforce_compute_snakes

__all__ = ["compute_snakes_multilevel"]


def compute_snakes(A, B, compare, rect=None):
    if rect is None:
        rect = (0, 0, len(A), len(B))
    i0, j0, i1, j1 = rect

    # TODO: Implement this using difflib or recursive Myers or simpler
    # non-recursive Myers algorithm without the middle snake stuff

    # snakes = [(i, j, n)]
    snakes = bruteforce_compute_snakes(A[i0:i1], B[j0:j1], compare)
    snakes = [(i+i0, j+j0, n) for (i, j, n) in snakes]

    assert all(compare(A[i+k], B[j+k]) for (i, j, n) in snakes for k in range(n)), (
        'Sanity check of multilevel comparison algorithm failed.')
    return snakes


def compute_snakes_multilevel(A, B, compares, rect=None, level=None):
    """Compute snakes using a multilevel multi-predicate algorithm.

    TODO: Document this algorithm.
    """
    if level is None:
        level = len(compares) - 1
    if rect is None:
        rect = (0, 0, len(A), len(B))

    # Compute initial set of coarse snakes
    compare = compares[level]
    snakes = compute_snakes(A, B, compare, rect)
    if level == 0:
        return snakes

    newsnakes = [(0, 0, 0)]
    i0, j0, i1, j1 = rect
    for snake in snakes + [(i1, j1, 0)]:
        i, j, n = snake
        if i > i0 and j > j0:
            # Recurse to compute snakes with less accurate
            # compare predicates between the coarse snakes
            subrect = (i0, j0, i, j)
            newsnakes += compute_snakes_multilevel(A, B, compares, subrect, level-1)
        if n > 0:
            li, lj, ln = newsnakes[-1]
            if li+ln == i and lj+ln == j:
                # Merge contiguous snakes
                newsnakes[-1] = (li, lj, ln + n)
            else:
                # Add new snake
                newsnakes.append(snake)
        i0 = i + n
        j0 = j + n
    # Pop empty snake from beginning if it wasn't extended inside the loop
    if newsnakes[0][2] == 0:
        newsnakes.pop(0)
    return newsnakes


def compute_diff_from_snakes(a, b, snakes, path="", config=None):
    "Compute diff from snakes."

    subpath = "/".join((path, "*"))
    diffit = config.differs[subpath]

    di = SequenceDiffBuilder()
    i0, j0, i1, j1 = 0, 0, len(a), len(b)
    for i, j, n in snakes + [(i1, j1, 0)]:
        if i > i0:
            di.removerange(i0, i-i0)
        if j > j0:
            di.addrange(i0, b[j0:j])

        for k in range(n):
            aval = a[i + k]
            bval = b[j + k]
            cd = diffit(aval, bval, path=subpath, config=config)
            if cd:
                di.patch(i + k, cd)

        # Update corner offsets for next rectangle
        i0, j0 = i+n, j+n
    return di.validated()



#!/usr/bin/env python
"""A simple differ for diffing directories containing notebooks.

Does not handle file moves/renames (yet).
"""

import os
import filecmp
from functools import partial

from nbdime.utils import EXPLICIT_MISSING_FILE

pjoin = os.path.join


def merge_two_dicts(x, y):
    """Given two dicts, merge them into a new dict as a shallow copy."""
    z = x.copy()
    z.update(y)
    return z


class dircmp(filecmp.dircmp, object):
    """
    Compare the content of dir1 and dir2. In contrast with filecmp.dircmp, this
    subclass compares the content of files with the same path.
    """

    def phase0(self):
        """
        Find the files on left and right for comparison.
        Ensure that we only consider notebook files or directories.
        """
        super(dircmp, self).phase0()
        self.left_list = list(filter(partial(ipynb_only, self.left), getattr(self, 'left_list')))
        self.right_list = list(filter(partial(ipynb_only, self.right), getattr(self, 'right_list')))

    def phase2(self):
        """
        Distinguish files, directories, funnies.
        Ensure files are only notebooks.
        """
        super(dircmp, self).phase2()
        self.common_files = list(filter(partial(ipynb_only, self.left), getattr(self, 'common_files')))

    def phase3(self):
        """
        Find out differences between common files.
        Ensure we are using content comparison with shallow=False.
        """
        fcomp = filecmp.cmpfiles(self.left, self.right, self.common_files,
                                 shallow=False)
        self.same_files, self.diff_files, self.funny_files = fcomp

    def phase4(self):
        """
        Find out differences between common subdirectories
        A new dircmp object is created for each common subdirectory,
        these are stored in a dictionary indexed by filename.
        The hide and ignore properties are inherited from the parent
        """

        self.subdirs = {}
        for x in self.common_dirs:
            a_x = os.path.join(self.left, x)
            b_x = os.path.join(self.right, x)
            self.subdirs[x] = self.__class__(a_x, b_x, self.ignore, self.hide)

    methodmap = merge_two_dicts(filecmp.dircmp.methodmap, dict(
                     subdirs=phase4,
                     same_files=phase3, diff_files=phase3, funny_files=phase3,
                     common_dirs = phase2, common_files=phase2, common_funny=phase2,
                     left_list=phase0, right_list=phase0))


def ipynb_only(parent, path):
    return os.path.isdir(pjoin(parent, path)) or os.path.splitext(path)[1] == '.ipynb'


def find_all_sub_notebooks(dirpath):
    for dirpath, _, filenames in os.walk(dirpath, topdown=False):
        for f in filenames:
            if os.path.splitext(f)[1] == '.ipynb':
                yield pjoin(dirpath, f)


def diff_directories(a, b, dc=None):
    """Iterate over differing files in two directories.

    Based on directory/file name only, so renames are handled as
    a full deletion + addition.

    Parameters:
    -----------
        a: First directory
        b: Second directory
        dc: Recursion variable (custom dircmp instance)
    """
    if dc is None:
        dc = dircmp(a, b, ignore=[])
    # Yield deleted files from current directory, or deleted directories:
    for deleted in dc.left_only:
        deleted = pjoin(dc.left, deleted)
        if os.path.isdir(deleted):
            for del_f in find_all_sub_notebooks(deleted):
                yield (del_f, EXPLICIT_MISSING_FILE)
        else:
            yield (deleted, EXPLICIT_MISSING_FILE)
    # Yield added files from current directory, or added directories:
    for added in dc.right_only:
        added = pjoin(dc.right, added)
        if os.path.isdir(added):
            for add_f in find_all_sub_notebooks(added):
                yield (EXPLICIT_MISSING_FILE, add_f)
        else:
            yield (EXPLICIT_MISSING_FILE, added)
    # Recurse for common directories:
    for _, sub_dc in dc.subdirs.items():
        for sub in diff_directories(None, None, sub_dc):
            yield sub

    # Yield changed files from current directory:
    for changed in dc.diff_files:
        yield (pjoin(dc.left, changed), pjoin(dc.right, changed))



#!/usr/bin/env python
"""A git mergetool plugin for notebooks.

Uses nbdime to merge notebooks instead of plain text merges of JSON.

Enable in your global git config with:

    git-nbmergetool config --enable [--global | --system]

Use with:

    git mergetool [<commit> [<commit>]]
"""
import sys
from subprocess import check_call, CalledProcessError

from nbdime.webapp import nbmergetool
from nbdime.args import add_generic_args, add_git_config_subcommand, ConfigBackedParser


def enable(scope=None, set_default=False):
    """Enable nbdime git mergetool"""
    cmd = ['git', 'config']
    if scope:
        cmd.append('--%s' % scope)

    # Register CLI tool
    check_call(cmd + ['mergetool.nbdime.cmd', 'git-nbmergetool merge "$BASE" "$LOCAL" "$REMOTE" "$MERGED"'])

    # Common setting:
    check_call(cmd + ['mergetool.prompt', 'false'])

    if set_default:
        # Set default tool to webapp
        check_call(cmd + ['merge.tool', 'nbdime'])


def disable(scope=None, _=None):
    """Disable nbdime git mergetool"""
    cmd = ['git', 'config']
    if scope:
        cmd.append('--%s' % scope)
    try:
        check_call(cmd + ['--unset', 'merge.tool'])
    except CalledProcessError:
        # already unset
        pass


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    import argparse
    parser = ConfigBackedParser('git-nbmergetool', description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    add_generic_args(parser)
    subparsers = parser.add_subparsers(dest='subcommand')

    merge_parser = subparsers.add_parser('merge',
        description="The actual entrypoint for the mergetool. Git will call this."
    )
    nbmergetool.build_arg_parser(merge_parser)

    config = add_git_config_subcommand(subparsers,
        enable, disable,
        subparser_help="Configure git to use nbdime via `git mergetool`",
        enable_help="enable nbdime mergetool via git config",
        disable_help="disable nbdime mergetool via git config")
    config.add_argument('--set-default', action='store_true', dest='set_default',
        help="set nbdime as default mergetool"
    )

    opts = parser.parse_args(args)
    if opts.subcommand == 'merge':
        return nbmergetool.main_parsed(opts)
    elif opts.subcommand == 'config':
        opts.config_func(opts.scope, opts.set_default)
        return 0
    else:
        parser.print_help()
        return 1


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from difflib import SequenceMatcher
from ..diff_format import SequenceDiffBuilder


__all__ = ["diff_sequence_difflib"]


def opcodes_to_diff(a, b, opcodes):
    "Convert difflib opcodes to nbdime diff format."
    di = SequenceDiffBuilder()
    for opcode in opcodes:
        action, abegin, aend, bbegin, bend = opcode
        asize = aend - abegin
        #bsize = bend - bbegin
        if action == "equal":
            # Unlike difflib we don't represent equal stretches explicitly
            pass
        elif action == "replace":
            di.removerange(abegin, asize)
            di.addrange(abegin, b[bbegin:bend])
        elif action == "insert":
            di.addrange(abegin, b[bbegin:bend])
        elif action == "delete":
            di.removerange(abegin, asize)
        else:
            raise RuntimeError("Unknown action {}".format(action))
    return di.validated()


def diff_sequence_difflib(a, b):
    """Compute the diff of two sequences.

    This implementation uses SequenceMatcher from the builtin Python difflib.

    By the difflib documentation this should work for sequences of
    hashable objects, i.e. the elements of the sequences are only
    compared for full equality. Therefore this function does not take
    a custom compare function like the other diff_sequence_* variants.
    """
    assert not any(isinstance(x, (list, dict)) for x in a), 'element in sequence not hashable'
    assert not any(isinstance(x, (list, dict)) for x in b), 'element in sequence not hashable'
    s = SequenceMatcher(None, a, b, autojunk=False)
    return opcodes_to_diff(a, b, s.get_opcodes())



#!/usr/bin/env python
# -*- coding:utf-8 -*-




import sys

from ..args import (
    ConfigBackedParser,
    add_generic_args, add_diff_args, add_web_args, add_filename_args,
    args_for_server, args_for_browse, process_diff_flags)
from .nbdimeserver import main_server as run_server
from .webutil import browse


# TODO: Tool server is passed a (mandatory?) single-use access token, which is
#       used to authenticate the browser session.

def build_arg_parser(parser=None):
    """
    Creates an argument parser for the diff tool, that also lets the
    user specify a port and displays a help message.
    """
    description = 'difftool for nbdime.'
    if parser is None:
        parser = ConfigBackedParser(
            description=description,
            add_help=True
            )
    add_generic_args(parser)
    add_web_args(parser, 0)
    add_diff_args(parser)
    add_filename_args(parser, ["local", "remote"])
    return parser


def main_parsed(opts):
    """Main function called after parsing CLI options

    Called by both main here and gitdifftool
    """
    process_diff_flags(opts)
    base = opts.local
    remote = opts.remote
    return run_server(
        difftool_args=dict(base=base, remote=remote),
        on_port=lambda port: browse(
            port=port,
            rel_url='difftool',
            **args_for_browse(opts)),
        **args_for_server(opts))


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    opts = build_arg_parser().parse_args(args)
    return main_parsed(opts)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import operator
from collections import defaultdict
import difflib

from ..diff_format import SequenceDiffBuilder, MappingDiffBuilder, validate_diff
from ..diff_utils import count_consumed_symbols

from .config import DiffConfig
from .sequences import diff_strings_linewise, diff_sequence
from .snakes import compute_snakes_multilevel, compute_diff_from_snakes

__all__ = ["diff"]


def default_predicates():
    return defaultdict(lambda: (operator.__eq__,))


def default_differs():
    return defaultdict(lambda: diff)


def compare_strings_approximate(x, y, threshold=0.7, maxlen=None):
    "Compare to strings with approximate heuristics."
    # TODO: Add configuration framework
    # TODO: Tune threshold with realistic sources

    # Fast cutoff when one is empty
    if bool(x) != bool(y):
        return False

    # Cutoff on equality: Python has fast hash functions for strings,
    # and lists of strings also works fine
    if len(x) == len(y) and x == y:
        return True

    # TODO: Investigate performance and quality of this difflib ratio approach,
    # possibly one of the weakest links of the notebook diffing algorithm.
    # Alternatives to try are the libraries diff-patch-match and Levenschtein

    # Informal benchmark normalized to operator ==:
    #    1.0  operator ==
    #  438.2  real_quick_ratio
    #  796.5  quick_ratio
    # 3088.2  ratio
    # The == cutoff will hit most of the time for long runs of
    # equal items, at least in the Myers diff algorithm.
    # Most other comparisons will likely not be very similar,
    # and the (real_)quick_ratio cutoffs will speed up those.

    # So the heavy ratio function is only used for close calls.
    # s = difflib.SequenceMatcher(lambda c: c in (" ", "\t"), x, y, autojunk=False)
    s = difflib.SequenceMatcher(None, x, y, autojunk=False)

    # Use only the fast ratio approximations first
    if s.real_quick_ratio() < threshold:
        return False
    if s.quick_ratio() < threshold:
        return False

    if maxlen is not None and len(x) > maxlen and len(y) > maxlen:
        # We know from above that there is not an exact similarity
        return False

    return s.ratio() > threshold


def diff(a, b, path="", config=None):
    "Compute the diff of two json-like objects, list or dict or string."

    if config is None:
        config = DiffConfig()

    if isinstance(a, list) and isinstance(b, list):
        d = diff_lists(a, b, path=path, config=config)
    elif isinstance(a, dict) and isinstance(b, dict):
        d = diff_dicts(a, b, path=path, config=config)
    elif isinstance(a, str) and isinstance(b, str):
        # Don't pass differs/predicates as the only possible use case is to
        # use a different character differ within each line or predicates
        # for comparing lines
        d = diff_strings_linewise(a, b)
    else:
        raise RuntimeError("Can currently only diff list, dict, or str objects.")

    # We can turn this off for performance after the library has been well tested:
    validate_diff(d)

    return d


def diff_string_lines(a, b, path="", config=None):
    """Diff two lists of strings (lines)"""

    # This is mainly about short-circuiting to avoid full snakes for equal content
    # since we know we can rely on __eq__ comparison
    if len(a) == len(b) and a == b:
        return []
    
    return diff_strings_linewise(a, b)


def diff_sequence_multilevel(a, b, path="", config=None):
    """Compute diff of two lists with configurable behaviour."""

    if config is None:
        config = DiffConfig()

    # Invoke multilevel snake computation algorithm
    compares = config.predicates[path or '/']
    snakes = compute_snakes_multilevel(a, b, compares)

    # Convert snakes to diff
    return compute_diff_from_snakes(a, b, snakes, path=path, config=config)


def diff_lists(a, b, path="", config=None, shallow_diff=None):
    """Compute diff of two lists with configurable behaviour."""

    if config is None:
        config = DiffConfig()

    # If multiple compares are provided to this path, delegate to multilevel algorithm
    compares = config.predicates[path or '/']
    if len(compares) > 1:
        assert shallow_diff is None
        return diff_sequence_multilevel(a, b, path=path, config=config)

    # First make a shallow sequence diff with custom compare,
    # unless it's provided for us
    if shallow_diff is None:
        shallow_diff = diff_sequence(a, b, compares[0])

    # Next we recurse to diff items in sequence that are considered
    # similar by compares[0] in the loop below
    subpath = "/".join((path, "*"))
    diffit = config.differs[subpath]

    # Count consumed items i,j from a,b, (i="take" in patch_list)
    i, j = 0, 0
    di = SequenceDiffBuilder()
    M = len(shallow_diff)
    for ie in range(M+1):
        if ie < M:
            # Consume n more unmentioned items before this diff entry
            # Note that index can be larger than i in the case where items
            # have been deleted from a and then insertions from b occur.
            e = shallow_diff[ie]
            index = e.key
            n = max(0, index - i)
            askip, bskip = count_consumed_symbols(e)
        else:
            # Consume final items after the last diff entry
            e = None
            n = len(a) - i
            askip, bskip = 0, 0
            assert n >= 0, "Sanity check failed: Cannot have negative remaining entries"
            assert len(b) - j == n, "Sanity check failed: Base/remote indexing mismatch"

        # Recursively diff the n items that have been deemed similar
        for k in range(n):
            aval = a[i + k]
            bval = b[j + k]
            if not config.is_atomic(aval, subpath):
                cd = diffit(aval, bval, path=subpath, config=config)
                if cd:
                    di.patch(i + k, cd)  # FIXME: Not covered in tests, create test situation

        # Keep count of consumed items
        i += n + askip
        j += n + bskip

        # Insert the diff entry from shallow diff unless past the end
        # (this either adds or removes items)
        if ie < M:
            di.append(e)

    # Sanity check
    assert i == len(a), "Sanity check failed: Did not process all entries in a"
    assert j == len(b), "Sanity check failed: Did not process all entries in b"

    return di.validated()


def diff_dicts(a, b, path="", config=None):
    """Compute diff of two dicts with configurable behaviour.

    Keys in both a and b will be handled based on

    Make a one-level diff of dicts a and b, using given compare
    operator to specify which items are considered the same.

    Items not mentioned in diff are items where compare(x, y) return True.
    For other items the diff will contain delete, insert, or replace entries.
    """
    if config is None:
        config = DiffConfig()

    if not isinstance(a, dict) or not isinstance(b, dict):
        raise TypeError('Arguments to diff_dicts need to be dicts, got %r and %r' % (a, b))
    akeys = set(a.keys())
    bkeys = set(b.keys())

    di = MappingDiffBuilder()

    # Sorting keys in loops to get a deterministic diff result
    for key in sorted(akeys - bkeys):
        di.remove(key)

    # Handle values for keys in both a and b
    for key in sorted(akeys & bkeys):
        avalue = a[key]
        bvalue = b[key]
        # If types are the same and nonatomic, recurse
        subpath = "/".join((path, key))
        if type(avalue) is type(bvalue) and not config.is_atomic(avalue, path=subpath):
            diffit = config.differs[subpath]
            dd = diffit(avalue, bvalue, path=subpath, config=config)
            if dd:
                di.patch(key, dd)
        else:
            if (path or '/') in config.predicates:
                # Could also this a warning, but I think it shouldn't be done
                raise RuntimeError(
                    "Found predicate(s) for path {} pointing to dict entry.".format(
                        path or '/'))
            if avalue != bvalue:
                di.replace(key, bvalue)

    for key in sorted(bkeys - akeys):
        di.add(key, b[key])

    return di.validated()



#!/usr/bin/env python
# -*- coding:utf-8 -*-




import sys

from ..args import (
    ConfigBackedParser,
    add_generic_args, add_diff_args, add_merge_args,
    add_web_args, add_filename_args, args_for_server, args_for_browse,
    process_diff_flags, Path)
from .nbdimeserver import main_server as run_server
from .webutil import browse


def build_arg_parser():
    """
    Creates an argument parser for the merge tool, that also lets the
    user specify a port and displays a help message.
    """
    description = 'Mergetool for Nbdime.'
    parser = ConfigBackedParser(
        description=description,
        add_help=True
        )
    add_generic_args(parser)
    add_diff_args(parser)
    add_merge_args(parser)
    add_web_args(parser, 0)
    parser.add_argument(
        "--no-base",
        dest="show_base",
        action="store_false",
        default=True,
        help="Don't display the base version."
    )
    add_filename_args(parser, ["base", "local", "remote"])
    parser.add_argument(
        '--out',
        type=Path,
        default=None,
        help="if supplied, the merged notebook is written "
             "to this file. Otherwise it cannot be saved.")
    return parser


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    arguments = build_arg_parser().parse_args(args)
    process_diff_flags(arguments)
    base = arguments.base
    local = arguments.local
    remote = arguments.remote
    output = arguments.out
    return run_server(
        outputfilename=output,
        on_port=lambda port: browse(
            port=port,
            rel_url='merge',
            base=base, local=local, remote=remote,
            **args_for_browse(arguments)),
        **args_for_server(arguments), 
        show_base=arguments.show_base,
    )


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding:utf-8 -*-

import io
import os
from collections import deque

os.environ['GIT_PYTHON_REFRESH'] = 'quiet'
from git import (
    Repo, InvalidGitRepositoryError, BadName, NoSuchPathError,
    GitCommandNotFound, Diffable
)

from nbdime.vcs.git.filter_integration import apply_possible_filter
from .utils import EXPLICIT_MISSING_FILE, pushd


# Git ref representing the working tree
GitRefWorkingTree = None

# Git ref representing the index
GitRefIndex = Diffable.Index


class BlobWrapper(io.StringIO):
    """StringIO with a name attribute"""
    name = ''


def get_repo(path):
    """Gets a Repo for path, also if path is a subdirectory of a repository

    Returns a tuple with the Repo object, and a list of subdirectories
    between path and the parent repository"""
    path = os.path.realpath(os.path.abspath(path))
    popped = deque()
    while True:
        try:
            repo = Repo(path)
            return (repo, tuple(popped))
        except (InvalidGitRepositoryError, NoSuchPathError):
            path, pop = os.path.split(path)
            if not pop:
                raise
            popped.appendleft(pop)


def find_repo_root(path):
    """Gets the directory containing the working tree"""
    repo = get_repo(path)[0]
    return repo.working_tree_dir


def traverse_tree(tree, subdirs):
    """Get the subtree according to given list of subdirectories"""
    if len(subdirs) == 0:
        return tree
    sub = subdirs[0]
    # Find first subtree whose name matches sub:
    subtree = next((st for st in tree.trees if st.name == sub), None)
    if subtree is None:
        raise KeyError('%s is not a subtree of %s' % (sub, tree.name))
    return traverse_tree(subtree, subdirs[1:])


def is_gitref(candidate):
    """Is candidate a gitref, or is it a file/filename?

    Returns false for collisions (e.g. when it is both a valid filename and gitref).
    Tests relative to current directory.
    """
    return (
        (candidate is None or not os.path.exists(candidate)) and
        candidate != EXPLICIT_MISSING_FILE and
        is_valid_gitref(candidate)
        )


def is_valid_gitref(ref, path=None):
    """Checks whether ref is a valid gitref in `path`, per git-rev-parse
    """
    try:
        repo = get_repo(path or os.curdir)[0]
        repo.commit(ref)
        return True
    except InvalidGitRepositoryError:
        return False
    except BadName:
        return False


def is_path_in_repo(path):
    """Checks whether path is part of a git repository"""
    try:
        get_repo(path)
        return True
    except InvalidGitRepositoryError:
        return False


def _get_diff_entry_stream(path, blob, ref_name, repo_dir):
    """Get a stream to the notebook, for a given diff entry's path and blob

    Returns None if path is not a Notebook file, and EXPLICIT_MISSING_FILE
    if path is missing, or the blob is None (unless diffing against working
    tree).
    """
    if path:
        if not path.endswith('.ipynb'):
            return None
        if ref_name is GitRefWorkingTree:
            # Diffing against working copy, use file on disk!
            with pushd(repo_dir):
                # We are diffing against working dir, so ensure we apply
                # any git filters before comparing:
                ret = apply_possible_filter(path)
                # ret == path means no filter was applied
                if ret != path:
                    return ret
                try:
                    return io.open(path, encoding='utf-8')
                except IOError:
                    return EXPLICIT_MISSING_FILE
        elif blob is None:
            # GitPython uses a None blob to indicate if a file was deleted or
            # added. Workaround for GitPython issue #749.
            return EXPLICIT_MISSING_FILE
        else:
            # There were strange issues with passing blob data_streams around,
            # so we solve this by reading into a StringIO buffer.
            # The penalty should be low as long as changed_notebooks are used
            # properly as an iterator.
            f = BlobWrapper(blob.data_stream.read().decode('utf-8'))
            f.name = '%s (%s)' % (
                path,
                ref_name if ref_name != GitRefIndex else '<INDEX>'
            )
            return f
    return EXPLICIT_MISSING_FILE


def changed_notebooks(ref_base, ref_remote, paths=None, repo_dir=None):
    """Iterator over all notebooks in path that has changed between the two git refs

    References are all valid values according to git-rev-parse, or one of
    the special sentinel values GitRefWorkingTree or GitRefIndex.

    Iterator value is a base/remote pair of streams to Notebooks
    or EXPLICIT_MISSING_FILE for added/removed files.
    """
    repo, popped = get_repo(repo_dir or os.curdir)
    if repo_dir is None:
        repo_dir = os.path.relpath(repo.working_tree_dir, os.curdir)
    if isinstance(paths, str):
        paths = (paths,)
    if paths and popped:
        # All paths need to be prepended by popped
        paths = [os.path.join(*(popped + (p,))) for p in paths]

    # Get tree/index for base
    if ref_base == GitRefIndex:
        tree_base = repo.index
    else:
        tree_base = repo.commit(ref_base).tree

    if ref_remote in (GitRefWorkingTree, GitRefIndex):
        diff = tree_base.diff(ref_remote, paths)
    else:
        # Get remote tree and diff against base:
        tree_remote = repo.commit(ref_remote).tree
        diff = tree_base.diff(tree_remote, paths)

    # Return the base/remote pair of Notebook file streams
    for entry in diff:
        fa = _get_diff_entry_stream(
            entry.a_path, entry.a_blob, ref_base, repo_dir)
        if fa is None:
            continue
        fb = _get_diff_entry_stream(
            entry.b_path, entry.b_blob, ref_remote, repo_dir)
        if fb is None:
            continue
        yield (fa, fb)



#!/usr/bin/env python
"""A git merge driver for notebooks.

Uses nbdime to merge notebooks instead of plain text merges of JSON.
Note that this requires the following to be set in .gitattributes to correctly
identify filetypes with the driver:

    *.ipynb     merge=jupyternotebook

Enable in your global git config with:

    git-nbmergedriver config --enable [--global | --system]

Use with:

    git merge [<commit> [<commit>]]
"""



import argparse
import io
import os
import sys
from subprocess import check_call, CalledProcessError

from nbdime import nbmergeapp
from nbdime.args import (
    add_generic_args, add_diff_args, add_merge_args, add_filename_args,
    add_git_config_subcommand, ConfigBackedParser,
)
from nbdime.utils import locate_gitattributes, ensure_dir_exists


def enable(scope=None):
    """Enable nbdime git merge driver"""
    cmd = ['git', 'config']
    if scope:
        cmd.append('--%s' % scope)

    check_call(cmd + ['merge.jupyternotebook.driver', 'git-nbmergedriver merge %O %A %B %L %P'])
    check_call(cmd + ['merge.jupyternotebook.name', 'jupyter notebook merge driver'])

    gitattributes = locate_gitattributes(scope)
    if gitattributes is None:
        assert scope is None, "No gitattributes found for scope: %s" % scope
        print("No .git directory in %s, skipping git attributes" % os.curdir, file=sys.stderr)
        return

    if os.path.exists(gitattributes):
        with io.open(gitattributes, encoding="utf8") as f:
            if 'merge=jupyternotebook' in f.read():
                # already written, nothing to do
                return
    else:
        ensure_dir_exists(os.path.dirname(gitattributes))

    with io.open(gitattributes, 'a', encoding="utf8") as f:
        f.write(u'\n*.ipynb\tmerge=jupyternotebook\n')


def disable(scope=None):
    """Disable nbdime git merge drivers"""
    cmd = ['git', 'config']
    if scope:
        cmd.append('--%s' % scope)
    try:
        check_call(cmd + ['--remove-section', 'merge.jupyternotebook'])
    except CalledProcessError:
        # already unset
        pass


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = ConfigBackedParser('git-nbmergedriver', description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest='subcommand')

    merge_parser = subparsers.add_parser('merge',
        description="The actual entrypoint for the merge tool. Git will call this."
    )
    add_generic_args(parser)
    add_diff_args(merge_parser)
    add_merge_args(merge_parser)

    # Argument list, we are given base, local, remote
    add_filename_args(merge_parser, ["base", "local", "remote"])

    # TODO: support git-config-specified conflict markers inside sources
    merge_parser.add_argument('marker')
    merge_parser.add_argument('out', nargs='?')
    # "The merge driver can learn the pathname in which the merged result will
    # be stored via placeholder %P"
    # - NOTE: This is not where the driver should store its output, see below!


    add_git_config_subcommand(
        subparsers,
        enable, disable,
        subparser_help="Configure git to use nbdime for notebooks in `git merge`",
        enable_help="enable nbdime merge driver via git config",
        disable_help="disable nbdime merge driver via git config")

    opts = parser.parse_args(args)
    if opts.subcommand == 'merge':
        # "The merge driver is expected to leave the result of the merge in the
        # file named with %A by overwriting it, and exit with zero status if it
        # managed to merge them cleanly, or non-zero if there were conflicts."
        opts.out = opts.local
        # mergeapp expects an additional decisions arg:
        opts.decisions = False
        return nbmergeapp.main_merge(opts)
    elif opts.subcommand == 'config':
        opts.config_func(opts.scope)
        return 0
    else:
        parser.print_help()
        return 1


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import io
import json
import os
import sys

import nbformat

from .args import ConfigBackedParser, Path, prettyprint_config_from_args
from .log import logger
from .merging import merge_notebooks
from .prettyprint import pretty_print_merge_decisions
from .utils import EXPLICIT_MISSING_FILE, read_notebook, setup_std_streams

_description = ('Merge two Jupyter notebooks "local" and "remote" with a '
                'common ancestor "base". If base is left out, it uses an '
                'empty notebook as the base.')


def main_merge(args):
    bfn = args.base
    lfn = args.local
    rfn = args.remote
    mfn = args.out

    from .args import process_diff_flags
    process_diff_flags(args)

    for fn in (bfn, lfn, rfn):
        if not os.path.exists(fn) and fn != EXPLICIT_MISSING_FILE:
            logger.error("Cannot find file '%s'", fn)
            return 1

    if lfn == rfn == EXPLICIT_MISSING_FILE:
        # Deleted both locally and remotely
        # Special case not well solved by routines below
        _handle_agreed_deletion(bfn, mfn, args)
        # Agreed on deletion = no conflics = return 0
        return 0

    # Git seems to give empty base file for double insertions
    b = read_notebook(bfn, on_null='minimal', on_empty='minimal')
    l = read_notebook(lfn, on_null='minimal')
    r = read_notebook(rfn, on_null='minimal')

    merged, decisions = merge_notebooks(b, l, r, args)
    conflicted = [d for d in decisions if d.conflict]

    returncode = 1 if conflicted else 0

    if conflicted:
        logger.warning("Conflicts occurred during merge operation.")
    else:
        logger.debug("Merge completed successfully with no unresolvable conflicts.")

    if args.decisions:
        if mfn:
            # write decisions as JSON file
            with io.open(mfn, "w", encoding="utf8") as outfile:
                json.dump(decisions, outfile, indent=2)
                outfile.write("\n")
        else:
            # Print merge decisions (including unconflicted)
            config = prettyprint_config_from_args(args, out=io.StringIO())
            pretty_print_merge_decisions(b, decisions, config=config)
            logger.warning("Decisions:\n%s", config.out.getvalue())
    elif mfn:
        # Write partial or fully completed merge to given foo.ipynb filename
        nbformat.write(merged, mfn)
        logger.info("Merge result written to %s", mfn)
    else:
        # Write merged notebook to terminal
        nbformat.write(merged, sys.stdout)
    return returncode


def _handle_agreed_deletion(base_fn, output_fn, args=None):
    """Handle merge when file has been deleted both locally and remotely"""
    assert base_fn != EXPLICIT_MISSING_FILE, (
        'sanity check failed: cannot have agreed decision on base %r' % base_fn)
    b = read_notebook(base_fn, on_null='minimal')
    if args and args.decisions:
        # Print merge decision (delete all)
        from .diffing.notebooks import diff_notebooks
        from .merging.decisions import MergeDecisionBuilder
        # Build diff for deleting all content:
        diff = diff_notebooks(b, {})
        # Make agreed decision from diff:
        bld = MergeDecisionBuilder()
        bld.agreement([], local_diff=diff, remote_diff=diff)
        decisions = bld.validated(b)
        # Print decition
        config = prettyprint_config_from_args(args, out=io.StringIO())
        pretty_print_merge_decisions(b, decisions, config=config)
        logger.warning("Decisions:\n%s", config.out.getvalue())

    elif output_fn:
        # Delete file if existing, if not do nothing
        if os.path.exists(output_fn):
            os.remove(output_fn)
            logger.info("Output file deleted: %s", output_fn)


def _build_arg_parser():
    """Creates an argument parser for the nbdiff command."""
    parser = ConfigBackedParser(
        description=_description,
        add_help=True,
        )
    from .args import (
        add_generic_args, add_diff_args, add_merge_args, filename_help,
        add_filename_args, add_prettyprint_args
    )
    add_generic_args(parser)
    add_diff_args(parser)
    add_merge_args(parser)
    add_prettyprint_args(parser)

    parser.add_argument(
        "base",
        type=Path,
        help=filename_help['base'],
        nargs='?',
        default=EXPLICIT_MISSING_FILE)
    add_filename_args(parser, ['local', 'remote'])

    parser.add_argument(
        '--out',
        default=None,
        type=Path,
        help="if supplied, the merged output is written "
             "to this file. Otherwise it is printed to the "
             "terminal.")
    parser.add_argument(
        '--decisions',
        action="store_true",
        help="print a human-readable summary of conflicted "
             "merge decisions instead of merging the notebook.")

    return parser


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    setup_std_streams()
    arguments = _build_arg_parser().parse_args(args)
    return main_merge(arguments)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import operator
from collections import defaultdict

from .config import DiffConfig
from .seq_difflib import diff_sequence_difflib
from .seq_bruteforce import diff_sequence_bruteforce
from .seq_myers import diff_sequence_myers

__all__ = ["diff_strings_by_char", "diff_sequence", "diff_strings_linewise"]


# TODO: Configuration framework?
# legal_diff_sequence_algorithms = ["bruteforce", "difflib", "myers"]
diff_sequence_algorithm = "bruteforce"


def diff_sequence(a, b, compare=operator.__eq__):
    """Compute a shallow diff of two sequences.

    I.e. these algorithms do not recursively diff elements of the sequences.

    This is a wrapper for alternative diff implementations.
    """
    if diff_sequence_algorithm == "difflib":
        if compare is not operator.__eq__:
            raise RuntimeError("Cannot use difflib with comparison other than ==.")
        return diff_sequence_difflib(a, b)
    elif diff_sequence_algorithm == "bruteforce":
        return diff_sequence_bruteforce(a, b, compare)
    elif diff_sequence_algorithm == "myers":
        return diff_sequence_myers(a, b, compare)
    else:
        raise RuntimeError("Unknown diff_sequence_algorithm {}.".format(diff_sequence_algorithm))


def diff_strings_by_char(a, b, path="", config=None):
    "Compute char-based diff of two strings."
    assert isinstance(a, str) and isinstance(b, str), (
        'Arguments need to be string types. Got %r and %r' % (a, b))
    if a == b:
        return []
    else:
        return diff_sequence_difflib(a, b)


def diff_strings_linewise(a, b):
    """Do a line-wise diff of two strings
    """
    assert isinstance(a, str) and isinstance(b, str), (
        'Arguments need to be string types. Got %r and %r' % (a, b))
    if a == b:
        return []
    lines_a = a.splitlines(True)
    lines_b = b.splitlines(True)

    from .generic import diff_lists, compare_strings_approximate
    config = DiffConfig(
        predicates=defaultdict(lambda: [
            compare_strings_approximate,
            operator.__eq__]),
        differs=defaultdict(lambda: diff_strings_by_char)
    )
    return diff_lists(lines_a, lines_b, config=config)



# coding: utf-8

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import sys
import json
import io

import nbformat

from .args import ConfigBackedParser
from .patching import patch_notebook
from .diff_utils import to_diffentry_dicts
from .utils import EXPLICIT_MISSING_FILE, read_notebook, setup_std_streams
from .prettyprint import pretty_print_notebook, PrettyPrintConfig


_description = "Apply patch from nbdiff to a Jupyter notebook."


def main_patch(args):
    base_filename = args.base
    patch_filename = args.patch
    output_filename = args.output

    for fn in (base_filename, patch_filename):
        if not os.path.exists(fn) and fn != EXPLICIT_MISSING_FILE:
            print("Missing file {}".format(fn))
            return 1

    before = read_notebook(base_filename, on_null='empty')
    with io.open(patch_filename, encoding="utf8") as patch_file:
        diff = json.load(patch_file)
    diff = to_diffentry_dicts(diff)

    after = patch_notebook(before, diff)

    if output_filename:
        nbformat.write(after, output_filename)
    else:
        try:
            nbformat.validate(after, version=4)
        except nbformat.ValidationError:
            print("Patch result is not a valid notebook, printing as JSON:")
            json.dump(after, sys.stdout)
        else:
            # This printer is to keep the unit tests passing,
            # some tests capture output with capsys which doesn't
            # pick up on sys.stdout.write()
            class Printer:
                def write(self, text):
                    print(text, end="")

            config = PrettyPrintConfig(out=Printer())

            pretty_print_notebook(after, config=config)

    return 0


def _build_arg_parser():
    """Creates an argument parser for the nbpatch command."""
    parser = ConfigBackedParser(
        description=_description,
        add_help=True,
        )
    from .args import add_generic_args, add_filename_args
    add_generic_args(parser)
    add_filename_args(parser, ["base", "patch"])
    parser.add_argument(
        '-o', '--output',
        default=None,
        help="if supplied, the patched notebook is written "
             "to this file. Otherwise it is printed to the "
             "terminal.")
    return parser


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    setup_std_streams()
    arguments = _build_arg_parser().parse_args(args)
    return main_patch(arguments)


if __name__ == "__main__":
    sys.exit(main())
